<?php

define('JSF_MAXBUF', 20);

class JSF_Parser
{
  private $name;
  public $color;
  private $now;
  public $state;
  public $styles;

  public $fg_default;
  public $bg_default;
  private $default_color_set;

  function __construct($syntax)
  {
    $this->name = $syntax;
    $this->color = Array();
    $this->state = Array();
    $this->now   = '';
    $this->fg_default = 0;
    $this->bg_default = 15;
    $this->default_color_set = false;
  }

  public function save()
  {
    return Array('state'  => $this->state,
                 'color'  => $this->color,
                 'styles' => $this->styles,
                 'bg_default' => $this->bg_default,
                 'fg_default' => $this->fg_default);
  }
  public function load($obj)
  {
    $this->color = $obj['color'];
    $this->state = $obj['state'];
    $this->styles= $obj['styles'];
    $this->bg_default = $obj['bg_default'];
    $this->fg_default = $obj['fg_default'];
  }

  private function ParseColorDecl($line)
  {
    $tok = preg_split("/[ \t]+/", $line);

    $fg    = $this->fg_default;
    $bg    = $this->bg_default;

    $under = false;
    $inv   = false;
    $bold  = 400;
    $italic= false;
    $blink = false;
    $name = '';
    foreach($tok as $tok)
    {
      if(!$name) $name = $tok;
      else switch($tok)
      {
        case 'bold':      $bold = 700; break;
        case 'semibold':  $bold = 600; break;
        case 'italic':$italic=true; break;
        case 'blink': $blink=true; break;
        case 'inverse': $inv = true; break;
        case 'underline': $under = true; break;
        case 'black':   $fg &= ~15; $fg |= 0; break;
        case 'blue':    $fg &= ~15; $fg |= 1; break;
        case 'green':   $fg &= ~15; $fg |= 2; break;
        case 'cyan':    $fg &= ~15; $fg |= 3; break;
        case 'red':case 'reg':
                        $fg &= ~15; $fg |= 4; break;
        case 'magenta': $fg &= ~15; $fg |= 5; break;
        case 'yellow':  $fg &= ~15; $fg |= 6; break;
        case 'white':   $fg &= ~15; $fg |= 7; break;
        case 'bg_black':   $bg &= ~15; $bg |= 0; break;
        case 'bg_blue':    $bg &= ~15; $bg |= 1; break;
        case 'bg_green':   $bg &= ~15; $bg |= 2; break;
        case 'bg_cyan':    $bg &= ~15; $bg |= 3; break;
        case 'bg_red':case 'bg_reg':
                           $bg &= ~15; $bg |= 4; break;
        case 'bg_magenta': $bg &= ~15; $bg |= 5; break;
        case 'bg_yellow':  $bg &= ~15; $bg |= 6; break;
        case 'bg_white':   $bg &= ~15; $bg |= 7; break;
        case 'BLACK':   $fg &= ~15; $fg |= 8+0; break;
        case 'BLUE':    $fg &= ~15; $fg |= 8+1; break;
        case 'GREEN':   $fg &= ~15; $fg |= 8+2; break;
        case 'CYAN':    $fg &= ~15; $fg |= 8+3; break;
        case 'RED':case 'REG':
                        $fg &= ~15; $fg |= 8+4; break;
        case 'MAGENTA': $fg &= ~15; $fg |= 8+5; break;
        case 'YELLOW':  $fg &= ~15; $fg |= 8+6; break;
        case 'WHITE':   $fg &= ~15; $fg |= 8+7; break;
        case 'BG_BLACK':   $bg &= ~15; $bg |= 8+0; break;
        case 'BG_BLUE':    $bg &= ~15; $bg |= 8+1; break;
        case 'BG_GREEN':   $bg &= ~15; $bg |= 8+2; break;
        case 'BG_CYAN':    $bg &= ~15; $bg |= 8+3; break;
        case 'BG_RED':case 'BG_REG':
                           $bg &= ~15; $bg |= 8+4; break;
        case 'BG_MAGENTA': $bg &= ~15; $bg |= 8+5; break;
        case 'BG_YELLOW':  $bg &= ~15; $bg |= 8+6; break;
        case 'BG_WHITE':   $bg &= ~15; $bg |= 8+7; break;
        default:
        {
          if(preg_match('@^bg_[0-5][0-5][0-5]@', $tok))
            $bg = 10000 + (int)substr($tok, 3);
          if(preg_match('@^fg_[0-5][0-5][0-5]@', $tok))
            $fg = 10000 + (int)substr($tok, 3);
        }
      }
    }

    $this->color[$name] =
     Array
     (
       'fg' => $fg,
       'bg' => $bg,
       'under' => $under,
       'inv'   => $inv,
       'bold'  => $bold,
       'blink' => $blink,
       'italic'=> $italic,
       'name' => $name
     );

    if(!$this->default_color_set)
    {
      $this->default_color_set = true;
      $this->fg_default = $fg;
      $this->bg_default = $bg;
    }
  }

  private function ParseStateStart($line)
  {
    $tok = preg_split("/[ \t]+/", $line);
    $name  = $tok[0];
    $state = $tok[1];

    unset($this->state[$this->now]['cur_table']);

    $this->state[$name] =
     Array
     (
       'name'      => $name,
       'color'     => $state,
       'cur_table' => false,
       'options'   => Array()
     );
    $this->now = $name;
  }

  private function ParseStateLine($tab)
  {
    $state = &$this->state[$this->now];

    $is_table = $state['cur_table'] !== false;

    #print "{this->now}: $is_table {$state['cur_table']} ".json_encode($tab)."\n";
    #print_r($state['options']);

    if($is_table)
    {
      $tableno = $state['cur_table'];

      if($tab == Array('done'))
      {
        $state['cur_table'] = false;
      }
      else
      {
        $table = &$state['options'][$tableno]['table'];

        $s = $tab[0];

        $s = preg_replace('@^"(.*)"$@', '\1', $s);
        $s = str_replace(Array('\n', '\r', '\t', '\\"'),
                         Array("\n", "\r", "\t", '"'), $s);
        $s = preg_replace('@\\\\([^\\\\])@', '\1', $s);
        $next_state = $tab[1];

        if($state['options'][$tableno]['strings'] == 2)
        {
          $s = strtoupper($s);
        }

        $table[$s] = $next_state;
      }
    }
    else
    {
      $regexp = false;
      $new_state = '';
      $noeat  = false;
      $buffer = false;
      $recolor= 0;
      $strings= '';
      foreach($tab as $item)
      {
        if($regexp === false)
        {
          $regexp = '';
          if($item != '*')
          {
            $s = $item;
            #print "in($item)";
            $s = preg_replace('@^"(.*)"$@', '\1', $s);
            $s = str_replace(Array('\n', '\r', '\t', '\"'),
                             Array("\n", "\r", "\t", '"'), $s);
            $s = preg_replace('@\\\\([^\\\\-])@', '\1', $s);
            $s = str_replace(']', '\]', $s);

            #print "out($s)\n";
            if(preg_match('@^\^@', $s)) $s = '\\'.$s;
            $regexp = '^['.$s.']$';
          }
          continue;
        }
        if(!$new_state)
        {
          $new_state = $item;
        }
        elseif($item == 'noeat')  $noeat = true;
        elseif($item == 'buffer') $buffer = true;
        elseif(preg_match('@^recolor=@', $item)) $recolor = (int)substr($item, 8);
        elseif($item == 'strings') $strings = 1;
        elseif($item == 'istrings') $strings = 2;
        else
        {
          // huh?
        }
      }
      if($regexp == '')
        $allowed_chars = true;
      else
      {
        $allowed_chars = Array();
        for($c=0; $c<256; ++$c)
          if(preg_match("$regexp", chr($c)))
            $allowed_chars[$c] = chr($c);
      }

      $choice = Array
      (
        'regexp'   => $regexp,
        'state'    => $new_state,
        'noeat'    => $noeat,
        'buffer'   => $buffer,
        'recolor'  => $recolor,
        'strings'  => $strings,
        'table'    => Array(),
        'allowed'  => $allowed_chars
      );
      $state['options'][] = $choice;

      if($strings)
      {
        $state['cur_table'] = count($state['options'])-1;
      }
    }
  }

  public static function EGAcolor($color)
  {
    if($color >= 10000)
    {
      $r = (int)($color/100)%10;
      $g = (int)($color/10)%10;
      $b = (int)($color/1)%10;
      return sprintf('#%X%X%X', $r*15/5, $g*15/5, $b*15/5);
    }

    static $EGA_COLOR_MAP = Array
    (
      '#000', '#33A', '#1A1', '#1AA',
      '#A11', '#A1A', '#A51', '#AAA',
      '#555', '#77F', '#5F5', '#5FF',
      '#F55', '#F5F', '#FF5', '#FFF'
    );

    return $EGA_COLOR_MAP[$color];
  }

  private function GenStyles()
  {
    $this->styles = Array();
    foreach($this->color as $name => &$color)
    {
      $defs = Array();
      /*
      ob_start();
       print_r($color);
       $tmp = ob_get_contents();
      ob_end_clean();
      $defs[] = "\n/"."* $tmp *"."/\n";
      */

      $defs[] = 'background:'.$this->EGAcolor($color[$color['inv']?'fg':'bg']);
      $defs[] = 'color:'.$this->EGAcolor($color[$color['inv']?'bg':'fg']);
      if($color['under']) $defs[] = 'text-decoration:underline';
      if($color['bold'] != 400) $defs[] = "font-weight:{$color['bold']}";
      if($color['italic']) $defs[] = 'font-style:italic';
      $style = '{'.join(';', $defs).'}';

      $color['style'] = $style;

      $name = '.' . $this->EncodeStyleName($name);
      $this->styles[$style][$name] = $name;
    }
  }

  public function GetBackgroundColorInCSSformat()
  {
    return
      'background-color:'.$this->EGAcolor($this->bg_default).
     ';color:'.$this->EGAcolor($this->fg_default);
  }

  public function GetStyles()
  {
    $styletext = '';
    foreach($this->styles as $style => $names)
      $styletext .= join(',', $names) . $style . "\n";
    return $styletext;
  }

  public function ParseSyntax($jsf_lines)
  {
    foreach($jsf_lines as $line)
    {
      $line = preg_replace("@[\r\n]+\$@", '', $line);
      $line = preg_replace('@^((?:"(?:\\\\.|[^"])"|[^#"])*)(#.*)?@', '\1', $line);
      $line = preg_replace("@[ \t]*\$@", '', $line);
      #print "$line\n";
      if(preg_match('@^=@', $line))
      {
        $this->ParseColorDecl(substr($line, 1));
      }
      elseif(preg_match('@^:@', $line))
      {
        $this->ParseStateStart(substr($line, 1));
      }
      elseif(preg_match("@^[ \t]@", $line) || $line == 'done')
      {
        #print "$line\n";
        preg_match_all('@"(?:\\\\"|[^"])+"|\*|[-=\w]+@', $line, $tab);
        $tab = $tab[0];
        #print_r($tab);
        $this->ParseStateLine($tab);
      }
    }
    $prev_state = &$this->state[$this->now];
    unset($prev_state['cur_table']);

    $this->GenStyles();
  }

  public function EncodeStyleName($name)
  {
    //return "jsf_{$this->name}_$name";
    //print_r($this->color[$name]);
    //$int = crc32($this->name.'.'.$name);
    $int = crc32($this->color[$name]['style']);

    $c="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_-"/*
       "".
       "".
       "".
       ""*/;

    $res='jSf';
    $n=strlen($c);
    if($int<0)$int=-$int;
    while($int>0)
    {
      $res .= $c[$int%$n];
      $int = (int)($int/$n);
    }
    return /*utf8_encode*/($res);
  }
};

class JSF_HTMLout
{
  private $cur_span;
  private $cur_color;
  private $pending_span;
  private $pending_color;
  private $text;

  function __construct()
  {
    $this->text         = '';
    $this->pending_span = false;
    $this->cur_span     = false;
  }
  function __destruct()
  {
    $this->Undo();
    print $this->text;
  }
  public function AppendText($text, $pending_span, &$pending_color)
  {
    if(!preg_match("@[^\t\r\n]@", $text))
    {
      $this->text .= $text;
      return;
    }
    if($pending_span != $this->cur_span)
    {
      $this->Undo();
      $this->cur_span = $pending_span;
      $this->cur_color= $pending_color;
      $this->text .= '<span class="'.htmlspecialchars($this->cur_span).'">';
      if($pending_color['blink'])
        $this->text .= '<blink>';
    }
    $this->text .= htmlspecialchars($text);
  }

  private function Undo()
  {
    if($this->cur_span !== false)
    {
      if($this->cur_color['blink']) $this->text .= '</blink>';
      $this->text .= '</span>';
    }
  }
};

class JSF_Renderer
{
  private $outputter;
  private $jsf_ref;

  private $buffer;
  private $head;
  private $tail;

  function __construct(JSF_Parser &$jsf_ref)
  {
    $this->outputter = new JSF_HTMLout;
    $this->jsf_ref = &$jsf_ref;

    $this->buffer = Array();
    $this->head   = 0;
    $this->tail   = 0;
  }

  function __destruct()
  {
    $this->Flush();
  }

  public function OutputChar($c, $color)
  {
    //print "*** out($c,$color)\n";

    $this->buffer[$this->head++] = Array($c, $color);
    $this->FlushUntil($this->head - JSF_MAXBUF);
  }

  public function Recolor($n, $color)
  {
    $reco = '';
    while($n > 0)
    {
      $ref = &$this->buffer[$this->head - $n--];
      $reco .= $ref[0];
      $ref[1] = $color;
    }
    //print "*** reco($reco,$color)\n";
  }

  public function Flush()
  {
    $this->FlushUntil($this->head);
  }

  private function FlushUntil($maxtail)
  {
    $buf = &$this->buffer;
    while($this->tail < $maxtail)
    {
      $ref = &$buf[$this->tail];

      $char  = $ref[0];
      $color = $ref[1];

      //print "*** actual_out($char,$color)\n";

      $this->outputter->AppendText($char,
         $this->jsf_ref->EncodeStyleName($color),
         $this->jsf_ref->color[$color]
                                  );

      unset($buf[$this->tail]);
      ++$this->tail;
    }
  }
};

class JSF_Applier
{
  private $jsf_ref;
  private $render;

  function __construct(JSF_Parser &$jsf)
  {
    $this->jsf_ref = &$jsf;
    $this->render  = new JSF_Renderer($jsf);
  }

  public function ApplyJSF($source_text)
  {
    $states = &$this->jsf_ref->state;

    reset($states);
    $first_state = key($states);

    $a=0;
    $b=strlen($source_text);

    $cur_state = $first_state;
    $buffer    = '';

    #print_r($states);

    $color = $states[$cur_state]['color'];

    if($a < $b)
    {
      $this->render->OutputChar($c = $source_text[$a], $color);
      while($a <= $b)
      {
        $state = &$states[$cur_state];

        $option_count = count($state['options']);
        for($o=$option_count; $o-- > 0; )
        {
          $option = &$state['options'][$o];
          $regexp = $option['regexp'];

          //print "state($cur_state)subj($c)test($regexp)goal({$option['state']})rec({$option['recolor']})";

          if($regexp == ''
          || preg_match("".$regexp."s", $c))
          {
            //print "bingo\n";
            break;
          }
          //print "nope\n";
        }

        $recolor   = $option['recolor'];
        $cur_state = $option['state'];

        #print "---\nState($cur_state),Character($c),Options:\n";
        #print_r($option);
        #print "--------\n";

        if($option['buffer']) $buffer='';

        #print_r($option);
        #print "c($c)buffer($buffer)\n";

        switch($option['strings'])
        {
          case 1:
            if(isset($option['table'][$buffer]))
            {
              $cur_state = $option['table'][$buffer];
              $recolor = strlen($buffer)+1;
            }
            break;

          case 2:
            $s = strtoupper($buffer);
            if(isset($option['table'][$s]))
            {
              $cur_state = $option['table'][$s];
              $recolor = strlen($buffer)+1;
            }
            break;
        }

        if($a >= $b && !$option['strings']) break;

        $buffer .= $c;
        if(strlen($buffer) > JSF_MAXBUF)
          $buffer = substr($buffer, -JSF_MAXBUF);

        $color = $states[$cur_state]['color'];
        if($recolor)
        {
          $this->render->Recolor(abs($recolor), $color);
        }

        if($a >= $b) break;

        if(!$option['noeat'])
        {
          ++$a;
          if($a < $b)
            $this->render->OutputChar($c = $source_text[$a], $color);
        }
      }
    }

    $this->render->Flush();
  }
};
